# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Gdk, Handy

from gettext import gettext as _

from lollypop.define import App, ScanType, NetworkAccessACL
from lollypop.widgets_row_device import DeviceRow
from lollypop.helper_passwords import PasswordsHelper


class SettingsDialog:
    """
        Dialog showing lollypop settings
    """

    __BOOLEAN = ["dark-ui", "artist-artwork", "auto-update", "background-mode",
                 "save-state", "import-playlists", "save-to-tags",
                 "show-compilations", "transitions", "transitions-party-only",
                 "recent-youtube-dl", "import-advanced-artist-tags",
                 "force-single-column", "hd-artwork", "network-access"]

    __RANGE = ["cover-size", "transitions-duration"]

    __COMBO = ["replay-gain", "orderby"]
    __ENTRY = ["invidious-server", "cs-api-key", "listenbrainz-user-token"]
    __LOCKED = ["data", "lyrics", "search", "playback"]

    def __init__(self):
        """
            Init dialog
        """
        self.__timeout_id = None
        self.__choosers = []
        self.__locked = []
        builder = Gtk.Builder()
        builder.add_from_resource("/org/gnome/Lollypop/SettingsDialog.ui")
        self.__settings_dialog = builder.get_object("settings_dialog")
        self.__progress = builder.get_object("progress")
        for setting in self.__BOOLEAN:
            button = builder.get_object("%s_boolean" % setting)
            value = App().settings.get_value(setting)
            button.set_state(value)
        for setting in self.__ENTRY:
            entry = builder.get_object("%s_entry" % setting)
            value = App().settings.get_value(setting).get_string()
            entry.set_text(value)
        for setting in self.__RANGE:
            widget = builder.get_object("%s_range" % setting)
            value = App().settings.get_value(setting).get_int32()
            widget.set_value(value)
        for setting in self.__COMBO:
            widget = builder.get_object("%s_combo" % setting)
            value = App().settings.get_enum(setting)
            widget.set_active(value)
        for locked in self.__LOCKED:
            widget = builder.get_object(locked)
            self.__locked.append(widget)
        self.__update_locked()
        self.__music_group = builder.get_object("music_group")
        for uri in App().settings.get_music_uris():
            button = self.__get_new_chooser(uri)
            self.__music_group.add(button)
        for device in App().settings.get_value("devices"):
            row = DeviceRow(device)
            builder.get_object("device_group").add(row)
        acl = App().settings.get_value("network-access-acl").get_int32()
        for key in NetworkAccessACL.keys():
            if acl & NetworkAccessACL[key]:
                builder.get_object("%s_button" % key).set_state(True)
        artists_count = App().artists.count()
        albums_count = App().albums.count()
        tracks_count = App().tracks.count()
        builder.get_object("stat_artists").set_title(
            _("Artists count: %s") % artists_count)
        builder.get_object("stat_albums").set_title(
            _("Albums count: %s") % albums_count)
        builder.get_object("stat_tracks").set_title(
            _("Tracks count: %s") % tracks_count)
        self.__settings_dialog.set_transient_for(App().window)
        self.__settings_dialog.connect("destroy", self.__on_destroy)
        passwords_helper = PasswordsHelper()
        passwords_helper.get("LASTFM", self.__on_get_password,
                             builder.get_object("lastfm_button"))
        passwords_helper.get("LIBREFM", self.__on_get_password,
                             builder.get_object("librefm_button"))
        builder.connect_signals(self)
        self.__controller = Gtk.EventControllerKey.new(self.__settings_dialog)
        self.__controller.connect("key-released", self.__on_key_released)
        if App().system_supports_color_schemes:
            builder.get_object("dark_row").hide()

    def show(self):
        """
            Show dialog
        """
        self.__settings_dialog.show()

#######################
# PROTECTED           #
#######################
    def _on_boolean_state_set(self, widget, state):
        """
            Save setting
            @param widget as Gtk.Switch
            @param state as bool
        """
        setting = widget.get_name()
        App().settings.set_value(setting,
                                 GLib.Variant("b", state))
        if setting == "dark-ui":
            if not App().player.is_party:
                settings = Gtk.Settings.get_default()
                settings.set_property("gtk-application-prefer-dark-theme",
                                      state)
        elif setting == "artist-artwork":
            App().window.container.reload_view()
        elif setting == "network-access":
            self.__update_locked()

    def _on_range_changed(self, widget):
        """
            Save value
            @param widget as Gtk.Range
        """
        setting = widget.get_name()
        value = widget.get_value()
        App().settings.set_value(setting, GLib.Variant("i", value))
        if setting == "cover-size":
            if self.__timeout_id is not None:
                GLib.source_remove(self.__timeout_id)
            self.__timeout_id = GLib.timeout_add(500,
                                                 self.__update_coversize,
                                                 widget)

    def _on_combo_changed(self, widget):
        """
            Save value
            @param widget as Gtk.ComboBoxText
        """
        setting = widget.get_name()
        value = widget.get_active()
        App().settings.set_enum(setting, value)
        if setting == "replay-gain":
            for plugin in App().player.plugins:
                plugin.build_audiofilter()
            App().player.reload_track()

    def _on_clean_artwork_cache_clicked(self, button):
        """
            Clean artwork cache
            @param button as Gtk.Button
        """
        App().task_helper.run(App().art.clean_all_cache)
        button.set_sensitive(False)

    def _on_google_api_key_changed(self, entry):
        """
            Save Key
            @param entry as Gtk.Entry
        """
        value = entry.get_text().strip()
        App().settings.set_value("cs-api-key", GLib.Variant("s", value))

    def _on_musicbrainz_api_key_changed(self, entry):
        """
            Save Key
            @param entry as Gtk.Entry
        """
        value = entry.get_text().strip()
        App().settings.set_value("listenbrainz-user-token",
                                 GLib.Variant("s", value))

    def _on_connect_button_clicked(self, button):
        """
            Connect to API
            @param button as Gtk.Button
        """
        def on_destroy(assistant):
            self.__settings_dialog.show()

        name = button.get_name()
        if button.get_tooltip_text():
            button.set_tooltip_text("")
            button.set_label(_("Connect"))
            App().ws_director.token_ws.clear_token(button.get_name(), True)
        else:
            if name == "GOOGLE":
                from lollypop.assistant_google import GoogleAssistant
                assistant = GoogleAssistant()
            elif name in ["LASTFM", "LIBREFM"]:
                from lollypop.assistant_lastfm import LastfmAssistant
                assistant = LastfmAssistant(name)
            else:
                from lollypop.assistant_musicbrainz import MusicbrainzAssistant
                assistant = MusicbrainzAssistant()
            assistant.set_transient_for(App().window)
            assistant.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
            assistant.show()
            assistant.connect("destroy", on_destroy)
            self.__settings_dialog.hide()
            button.set_label(_("Disconnect"))

    def _on_invidious_entry_changed(self, entry):
        """
            Update invidious server setting
            @param entry as Gtk.entry
        """
        uri = entry.get_text()
        App().settings.set_value("invidious-server", GLib.Variant("s", uri))

    def _on_acl_state_set(self, widget, state):
        """
            Save network acl state
            @param widget as Gtk.Switch
            @param state as bool
        """
        key = widget.get_name()
        acl = App().settings.get_value("network-access-acl").get_int32()
        if state:
            acl |= NetworkAccessACL[key]
        else:
            acl &= ~NetworkAccessACL[key]
        acl = App().settings.set_value("network-access-acl",
                                       GLib.Variant("i", acl))

    def _on_add_button_clicked(self, button):
        """
            Add a new chooser
            @param button as Gtk.Button
        """
        button = self.__get_new_chooser(None)
        self.__music_group.add(button)

    def _on_reset_button_clicked(self, button):
        """
            Reset database
            @param button as Gtk.Button
        """
        if button.get_style_context().has_class("red"):
            button.set_label(_("Reset collection"))
            button.set_sensitive(False)
            button.get_style_context().remove_class("red")
            App().scanner.reset_database()
        else:
            button.get_style_context().add_class("red")
            button.set_label(_("Are you sure?"))

#######################
# PRIVATE             #
#######################
    def __update_locked(self):
        """
            Update locked widgets
        """
        network_access = App().settings.get_value("network-access")
        for widget in self.__locked:
            widget.set_sensitive(network_access)

    def __get_new_chooser(self, uri):
        """
            Get a new chooser
            @param uri as str
            @return Handy.ActionRow
        """
        chooser = Gtk.FileChooserButton()
        chooser.show()
        chooser.set_local_only(False)
        chooser.set_action(Gtk.FileChooserAction.SELECT_FOLDER)
        chooser.set_valign(Gtk.Align.CENTER)
        chooser.set_hexpand(True)
        self.__choosers.append(chooser)
        if uri is not None:
            chooser.set_uri(uri)
        button = Gtk.Button.new_from_icon_name("list-remove-symbolic",
                                               Gtk.IconSize.BUTTON)
        button.show()
        button.set_valign(Gtk.Align.CENTER)
        row = Handy.ActionRow()
        row.show()
        row.add(chooser)
        row.add(button)
        button.connect("clicked", lambda x: self.__choosers.remove(chooser))
        button.connect("clicked", lambda x: row.destroy())
        return row

    def __update_coversize(self, widget):
        """
            Update cover size
            @param widget as Gtk.Range
        """
        self.__timeout_id = None
        App().task_helper.run(App().art.clean_all_cache)
        App().art.update_art_size()
        App().window.container.reload_view()

    def __on_destroy(self, widget):
        """
            Save settings and update if needed
            @param widget as Gtk.Window
        """
        # Music uris
        uris = []
        default = GLib.get_user_special_dir(
            GLib.UserDirectory.DIRECTORY_MUSIC)
        for chooser in self.__choosers:
            uri = chooser.get_uri()
            if uri is not None and uri not in uris:
                uris.append(uri)
        if not uris and default is not None:
            uris.append(GLib.filename_to_uri(default))

        previous = App().settings.get_value("music-uris")
        App().settings.set_value("music-uris", GLib.Variant("as", uris))

        if set(previous) != set(uris):
            to_delete = [uri for uri in previous if uri not in uris]
            to_scan = [uri for uri in uris if uri not in previous]
            if to_delete:
                App().scanner.update(ScanType.FULL)
            elif to_scan:
                App().scanner.update(ScanType.NEW_FILES, to_scan)

    def __on_get_password(self, attributes, password, service, button):
        """
            Set button state
            @param attributes as {}
            @param password as str
            @param service as str
            @param button as Gtk.Button
        """
        if attributes is not None:
            button.set_label(_("Disconnect"))
            button.set_tooltip_text("%s:%s" % (attributes["login"], password))

    def __on_key_released(self, event_controller, keyval, keycode, state):
        """
            Quit on escape
            @param event_controller as Gtk.EventController
            @param keyval as int
            @param keycode as int
            @param state as Gdk.ModifierType
        """
        if keyval == Gdk.KEY_Escape:
            self.__settings_dialog.destroy()



# Copyright (c) 2018 Philipp Wolfer <ph.wolfer@gmail.com>
# Copyright (c) 2018 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Soup, GLib, GObject, Gio

import json
from pickle import load, dump

from lollypop.logger import Logger
from lollypop.define import App, LOLLYPOP_DATA_PATH, Type
from lollypop.utils import get_network_available


class ListenBrainzWebService(GObject.GObject):
    """
        Submit listens to ListenBrainz.org.

        See https://listenbrainz.readthedocs.io/en/latest/dev/api.html
    """

    user_token = GObject.Property(type=str, default="plop")

    def __init__(self):
        """
            Init ListenBrainz object
        """
        GObject.GObject.__init__(self)
        try:
            self.__uri = "https://api.listenbrainz.org/1/submit-listens"
            self.__name = "listenbrainz"
            self.__queue = []
            self.start()
        except Exception as e:
            Logger.info("LastFM::__init__(): %s", e)

    def start(self):
        """
            Start web service (load save queue)
        """
        try:
            self.__cancellable = Gio.Cancellable()
            self.__queue = load(
                open(LOLLYPOP_DATA_PATH + "/%s_queue.bin" % self.__name, "rb"))
        except Exception as e:
            Logger.info("ListenBrainzWebService::start(): %s", e)
            self.__queue = []

    def stop(self):
        """
            Stop current tasks and save queue to disk
            @return bool
        """
        self.__cancellable.cancel()
        try:
            with open(LOLLYPOP_DATA_PATH + "/%s_queue.bin" % self.__name,
                      "wb") as f:
                dump(list(self.__queue), f)
        except Exception as e:
            Logger.info("ListenBrainzWebService::stop: %s", e)
        return True

    def listen(self, track, timestamp):
        """
            Submit a listen for a track (scrobble)
            @param track as Track
            @param timestamp as int
        """
        monitor = Gio.NetworkMonitor.get_default()
        if not App().settings.get_value(
                "listenbrainz-user-token").get_string():
            return
        elif App().settings.get_value("disable-scrobbling") or\
                not get_network_available() or\
                monitor.get_network_metered():
            self.__queue.append((track, timestamp))
        elif track.id is not None and track.id >= 0:
            App().task_helper.run(self.__listen, track, timestamp)

    def playing_now(self, track):
        """
            Submit a playing now notification for a track
            @param track as Track
        """
        if not App().settings.get_value(
                "listenbrainz-user-token").get_string():
            return
        elif App().settings.get_value("disable-scrobbling") or\
                not get_network_available():
            return
        if track.id is not None and track.id >= 0:
            App().task_helper.run(self.__playing_now, track)

    def love(self, artist, title):
        pass

    def unlove(self, artist, title):
        pass

    def set_loved(self, track, loved):
        pass

#######################
# PRIVATE             #
#######################
    def __listen(self, track, timestamp):
        """
            Scrobble track
            @param track as Track
            @param timestamp as int
        """
        tracks = self.__queue + [(track, timestamp)]
        self.__queue = []
        try:
            for (track, timestamp) in tracks:
                payload = self.__get_payload(track)
                payload[0]["listened_at"] = timestamp
                post_data = {
                    "listen_type": "single",
                    "payload": payload
                }
                data = self.__post_request(post_data)
                if data is not None:
                    Logger.debug("%s: %s", self.__uri, data)
                else:
                    self.__queue.append((track, timestamp))
        except Exception as e:
            Logger.error("ListenBrainzWebService::__listen(): %s" % e)

    def __playing_now(self, track):
        """
            Now playing track
            @param track as Track
        """
        try:
            payload = self.__get_payload(track)
            post_data = {
                "listen_type": "playing_now",
                "payload": payload
            }
            data = self.__post_request(post_data)
            if data is not None:
                Logger.debug("%s: %s", self.__uri, data)
        except Exception as e:
            Logger.error("ListenBrainzWebService::__playing_now(): %s" % e)

    def __post_request(self, data):
        msg = Soup.Message.new("POST", self.__uri)
        body = GLib.Bytes.new(json.dumps(data).encode("utf-8"))
        msg.set_request_body_from_bytes("application/json", body)
        request_headers = msg.get_property("request-headers")
        request_headers.append("Accept-Charset", "utf-8")
        request_headers.append("Authorization", "Token %s" % self.user_token)
        return App().task_helper.send_message_sync(msg, self.__cancellable)

    def __get_payload(self, track):
        """
            Build payload from track
            @param track as Track
            @return payload as []
        """
        if track.album.artist_ids[0] == Type.COMPILATIONS:
            artist = track.artists[0]
        else:
            artist = track.album.artists[0]
        payload = {
            "track_metadata": {
                "artist_name": artist,
                "track_name": track.title,
                "release_name": track.album_name,
                "additional_info": {
                    "media_player": "Lollypop",
                    "media_player_version": App().version,
                    "artist_mbids": [
                        mbid for mbid in track.mb_artist_ids if mbid
                    ],
                    "release_mbid": track.album.mb_album_id,
                    "recording_mbid": track.mb_track_id,
                    "tracknumber": track.number,
                    "duration_ms": track.duration
                }
            }
        }
        return [payload]



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib

from gettext import gettext as _
from hashlib import sha256

from lollypop.define import Type, App, SelectionListMask
from lollypop.shown import ShownLists, ShownPlaylists
from lollypop.utils import get_icon_name


class SelectionListRowMenu(Gio.Menu):
    """
        A menu related to a selection list row
    """
    def __init__(self, rowid, header=False):
        """
            Init menu
            @param rowid as int
            @param header as bool
        """
        Gio.Menu.__init__(self)
        if header:
            from lollypop.menu_header import MenuHeader
            label = ShownLists.IDS[rowid]
            icon_name = get_icon_name(rowid)
            self.append_item(MenuHeader(label, icon_name))
        startup_menu = Gio.Menu()
        selected = rowid == App().settings.get_value(
            "startup-id").get_int32()
        action = Gio.SimpleAction.new_stateful(
                                   "default_selection_id",
                                   None,
                                   GLib.Variant.new_boolean(selected))
        App().add_action(action)
        action.connect("change-state",
                       self.__on_default_change_state,
                       rowid)
        item = Gio.MenuItem.new(_("Default on startup"),
                                "app.default_selection_id")
        startup_menu.append_item(item)
        self.append_section(_("Startup"), startup_menu)

#######################
# PRIVATE             #
#######################
    def __on_default_change_state(self, action, variant, rowid):
        """
            Add to playlists
            @param action as Gio.SimpleAction
            @param variant as GVariant
            @param rowid as int
        """
        action.set_state(variant)
        if variant:
            App().settings.set_value("startup-id",
                                     GLib.Variant("i", rowid))
        else:
            App().settings.set_value("startup-id",
                                     GLib.Variant("i", -1))


class SelectionListMenu(Gio.Menu):
    """
        A menu for configuring SelectionList
    """

    def __init__(self, widget, mask, header=False):
        """
            Init menu
            @param widget as Gtk.Widget
            @param mask as SelectionListMask
            @param header as bool
        """
        Gio.Menu.__init__(self)
        self.__widget = widget
        self.__mask = mask

        if header:
            from lollypop.menu_header import MenuHeader
            if mask & SelectionListMask.PLAYLISTS:
                label = _("Playlists")
                icon_name = "folder-documents-symbolic"
            else:
                label = _("Sidebar")
                icon_name = "org.gnome.Lollypop-sidebar-symbolic"
            self.append_item(MenuHeader(label, icon_name))

        # Options
        if not App().window.folded and\
                not mask & SelectionListMask.PLAYLISTS:
            options_menu = Gio.Menu()
            action = Gio.SimpleAction.new_stateful(
                    "show_label",
                    None,
                    App().settings.get_value("show-sidebar-labels"))
            action.connect("change-state",
                           self.__on_show_label_change_state)
            App().add_action(action)
            options_menu.append(_("Show text"), "app.show_label")
            self.append_section(_("Options"), options_menu)
        # Shown menu
        shown_menu = Gio.Menu()
        if mask & SelectionListMask.PLAYLISTS:
            lists = ShownPlaylists.get(True)
            wanted = App().settings.get_value("shown-playlists")
        else:
            mask |= SelectionListMask.COMPILATIONS
            lists = ShownLists.get(mask, True)
            wanted = App().settings.get_value("shown-album-lists")
        for item in lists:
            if item[0] == Type.SEPARATOR:
                continue
            exists = item[0] in wanted
            encoded = sha256(item[1].encode("utf-8")).hexdigest()
            action = Gio.SimpleAction.new_stateful(
                encoded,
                None,
                GLib.Variant.new_boolean(exists))
            action.connect("change-state",
                           self.__on_shown_change_state,
                           item[0])
            App().add_action(action)
            shown_menu.append(item[1], "app.%s" % encoded)
        # Translators: shown => items
        self.append_section(_("Sections"), shown_menu)

#######################
# PRIVATE             #
#######################
    def __on_shown_change_state(self, action, variant, rowid):
        """
            Set action value
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
            @param rowid as int
        """
        action.set_state(variant)
        if self.__mask & SelectionListMask.PLAYLISTS:
            option = "shown-playlists"
        else:
            option = "shown-album-lists"
        wanted = list(App().settings.get_value(option))
        if variant and rowid not in wanted:
            wanted.append(rowid)
        elif rowid in wanted:
            wanted.remove(rowid)
        App().settings.set_value(option, GLib.Variant("ai", wanted))
        if self.__mask & SelectionListMask.PLAYLISTS:
            items = ShownPlaylists.get(True)
        else:
            items = ShownLists.get(self.__mask, True)
        if variant:
            for item in items:
                if item[0] == rowid:
                    self.__widget.add_value(item)
                    break
        else:
            self.__widget.remove_value(rowid)
            if self.__mask & SelectionListMask.SIDEBAR:
                startup_id = App().settings.get_value("startup-id")
                if startup_id == rowid:
                    App().settings.set_value("startup-id",
                                             GLib.Variant("i", -1))

    def __on_show_label_change_state(self, action, variant):
        """
            Update option
            @param action as Gio.SimpleAction
            @param variant as GVariant
        """
        action.set_state(variant)
        App().settings.set_value("show-sidebar-labels",
                                 GLib.Variant("b", variant))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Pango, Handy


from lollypop.define import App


class DeviceRow(Handy.ActionRow):
    """
        A device row
    """

    def __init__(self, name):
        """
            Init row
            @param name as str
        """
        Handy.ActionRow.__init__(self)
        self.__name = name
        label = Gtk.Label.new(name)
        label.set_property("halign", Gtk.Align.START)
        label.set_hexpand(True)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.show()
        button = Gtk.Button.new_from_icon_name("user-trash-symbolic",
                                               Gtk.IconSize.BUTTON)
        button.connect("clicked", self.__on_button_clicked)
        button.get_style_context().add_class("menu-button")
        button.show()
        button.set_valign(Gtk.Align.CENTER)
        self.add(label)
        self.add(button)
        self.show()

#######################
# PRIVATE             #
#######################
    def __on_button_clicked(self, button):
        """
            Remove device
            @param button as Gtk.Button
        """
        if button.get_image().get_style_context().has_class("red"):
            devices = list(App().settings.get_value("devices"))
            if self.__name in devices:
                index = devices.index(self.__name)
                devices.remove(self.__name)
                App().settings.set_value("devices",
                                         GLib.Variant("as", devices))
                App().albums.remove_device(index + 1)
                App().playlists.remove_device(index + 1)
            self.destroy()
        else:
            button.get_image().get_style_context().add_class("red")



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.define import ViewType


class PlaylistsContainer:
    """
        Playlists management for main view
    """

    def __init__(self):
        """
            Init container
        """
        pass

    def show_smart_playlist_editor(self, playlist_id):
        """
            Show a view allowing user to edit smart view
            @param playlist_id as int
        """
        view_type = ViewType.SCROLLED
        from lollypop.view_playlist_smart import SmartPlaylistView
        view = SmartPlaylistView(playlist_id, view_type)
        view.populate()
        view.show()
        self._stack.add(view)
        self._stack.set_visible_child(view)

##############
# PROTECTED  #
##############

############
# PRIVATE  #
############



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gtk

from gettext import gettext as _

from lollypop.define import App
from lollypop.helper_signals import SignalsHelper, signals


class PlayButton(Gtk.Button):
    """
        Gtk Button with a spinner
    """

    def __init__(self):
        """
            Init Button
        """
        Gtk.Button.__init__(self)
        self.__spinner = Gtk.Spinner.new()
        self.__spinner.show()
        self.__image = Gtk.Image.new_from_icon_name(
            "media-playback-start-symbolic", Gtk.IconSize.BUTTON)
        self.__image.show()
        self.set_image(self.__image)

    def set_loading(self, status):
        """
            Set button loading status
            @param status as bool
        """
        if status:
            self.set_image(self.__spinner)
            self.__spinner.start()
        else:
            self.set_image(self.__image)
            self.__spinner.stop()

    @property
    def image(self):
        """
            Get image
            @return Gtk.Image
        """
        return self.__image


class ButtonsPlayerWidget(Gtk.Box, SignalsHelper):
    """
        Box with playback buttons based on current player state
    """

    @signals
    def __init__(self, styles=[]):
        """
            Init box
            @param style as [str]
        """
        Gtk.Box.__init__(self)
        self.__prev_button_timeout_id = None
        self.__prev_button = Gtk.Button.new_from_icon_name(
            "media-skip-backward-symbolic", Gtk.IconSize.BUTTON)
        self.__prev_button.show()
        self.__prev_button.connect("clicked", self.__on_prev_button_clicked)
        self.__play_button = PlayButton()
        self.__play_button.show()
        self.__play_button.connect("clicked", self.__on_play_button_clicked)
        self.__next_button = Gtk.Button.new_from_icon_name(
            "media-skip-forward-symbolic", Gtk.IconSize.BUTTON)
        self.__next_button.show()
        self.__next_button.connect("clicked", self.__on_next_button_clicked)
        self.__play_button.set_sensitive(False)
        self.__prev_button.set_sensitive(False)
        self.__next_button.set_sensitive(False)
        self.get_style_context().add_class("linked")
        self.pack_start(self.__prev_button, True, True, 0)
        self.pack_start(self.__play_button, True, True, 0)
        self.pack_start(self.__next_button, True, True, 0)
        self.connect("destroy", self.__on_destroy)
        if styles:
            self.__set_styles(self.__prev_button, styles)
            self.__set_styles(self.__play_button, styles)
            self.__set_styles(self.__next_button, styles)
        else:
            self.__prev_button.set_size_request(42, -1)
            self.__play_button.set_size_request(60, -1)
            self.__next_button.set_size_request(42, -1)

        return [
            (App().player, "current-changed", "_on_current_changed"),
            (App().player, "status-changed", "_on_status_changed"),
            (App().player, "loading-changed", "_on_loading_changed"),
            (App().player, "next-changed", "_on_next_changed"),
            (App().player, "prev-changed", "_on_prev_changed")
        ]

    def update(self):
        """
            Update buttons
        """
        player = App().player
        self._on_current_changed(player)
        self._on_status_changed(player)

#######################
# PROTECTED           #
#######################
    def _on_current_changed(self, player):
        """
            Update toolbar
            @param player as Player
        """
        def update_button():
            self.__prev_button_timeout_id = None
            self.__prev_button.get_image().set_from_icon_name(
                "media-seek-backward-symbolic", Gtk.IconSize.BUTTON)

        if self.__prev_button_timeout_id is not None:
            GLib.source_remove(self.__prev_button_timeout_id)
        self.__prev_button.get_image().set_from_icon_name(
            "media-skip-backward-symbolic", Gtk.IconSize.BUTTON)
        self.__prev_button_timeout_id = GLib.timeout_add(
            App().settings.get_value("previous-threshold").get_int32(),
            update_button)

        if App().player.current_track.id is None:
            self.__play_button.set_sensitive(False)
            self.__prev_button.set_sensitive(False)
            self.__next_button.set_sensitive(False)
        else:
            self.__play_button.set_sensitive(True)
            self.__prev_button.set_sensitive(True)
            self.__next_button.set_sensitive(True)

    def _on_prev_changed(self, player):
        """
            Update prev button
            @param player as Player
        """
        if player.prev_track.id is not None:
            prev_artists = GLib.markup_escape_text(
                ", ".join(player.prev_track.artists))
            prev_title = GLib.markup_escape_text(player.prev_track.title)
            self.__prev_button.set_tooltip_markup("<b>%s</b> - %s" %
                                                  (prev_artists,
                                                   prev_title))
        else:
            self.__prev_button.set_tooltip_text("")

    def _on_next_changed(self, player):
        """
            Update toolbar
            @param player as Player
        """
        if player.next_track.id is not None:
            next_artists = GLib.markup_escape_text(
                ", ".join(player.next_track.artists))
            next_title = GLib.markup_escape_text(player.next_track.title)
            self.__next_button.set_tooltip_markup("<b>%s</b> - %s" %
                                                  (next_artists,
                                                   next_title))
        else:
            self.__next_button.set_tooltip_text("")

    def _on_status_changed(self, player):
        """
            Update play button state
            @param player as Player
        """
        if player.is_playing:
            self.__play_button.image.set_from_icon_name(
                "media-playback-pause-symbolic", Gtk.IconSize.BUTTON)
            self.__play_button.set_tooltip_text(_("Pause"))
        else:
            self.__play_button.image.set_from_icon_name(
                "media-playback-start-symbolic", Gtk.IconSize.BUTTON)
            self.__play_button.set_tooltip_text(_("Play"))

    def _on_loading_changed(self, player, status, track):
        """
            Show a spinner on play button
            @param player as Player
            @param status as bool
            @param track as Track
        """
        self.__play_button.set_loading(status)

#######################
# PRIVATE             #
#######################
    def __set_styles(self, widget, styles):
        """
            Add styles to widget
            @param widget as Gtk.Widget
            @param styles as [str]
        """
        context = widget.get_style_context()
        for style in styles:
            context.add_class(style)

    def __on_prev_button_clicked(self, button):
        """
            Previous track on prev button clicked
            @param button as Gtk.Button
        """
        App().player.prev()

    def __on_play_button_clicked(self, button):
        """
            Play/Pause on play button clicked
            @param button as Gtk.Button
        """
        if App().player.is_playing:
            App().player.pause()
        else:
            App().player.play()

    def __on_next_button_clicked(self, button):
        """
            Next track on next button clicked
            @param button as Gtk.Button
        """
        App().player.next()

    def __on_destroy(self, widget):
        """
            Stop timeout
            @param widget as Gtk.Widget
        """
        if self.__prev_button_timeout_id is not None:
            GLib.source_remove(self.__prev_button_timeout_id)
            self.__prev_button_timeout_id = None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject, Gtk

from lollypop.logger import Logger
from lollypop.container_history import HistoryContainer
from lollypop.utils import emit_signal


class StackContainer(Gtk.Stack):
    """
        A Gtk.Stack handling navigation
    """

    __gsignals__ = {
        "history-changed":   (GObject.SignalFlags.RUN_FIRST, None, ()),
        "set-sidebar-id":    (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        "set-selection-ids": (GObject.SignalFlags.RUN_FIRST, None,
                              (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self):
        """
            Init stack
        """
        Gtk.Stack.__init__(self)
        self.get_style_context().add_class("view")
        self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.set_hexpand(True)
        self.set_vexpand(True)
        self.__history = HistoryContainer()

    def add(self, view):
        """
            Add view to stack
            @param view as View
        """
        if view not in self.get_children():
            Gtk.Stack.add(self, view)

    def set_visible_child(self, view):
        """
            Set visible child in stack
            @param view as View
        """
        visible_child = self.get_visible_child()
        if visible_child == view:
            return
        elif visible_child is not None:
            visible_child.pause()
            if visible_child.__class__ != view.__class__ or\
                    visible_child.args != view.args:
                self.__history.add_view(visible_child)
            else:
                visible_child.destroy_later()
            Gtk.Stack.set_visible_child(self, view)
            emit_signal(self, "history-changed")
        else:
            Gtk.Stack.set_visible_child(self, view)

    def go_back(self):
        """
            Go back in stack
        """
        if self.__history:
            visible_child = self.get_visible_child()
            view = self.__history.pop()
            if view is not None:
                if view not in self.get_children():
                    self.add(view)
                Gtk.Stack.set_visible_child(self, view)
                if view.sidebar_id is not None:
                    emit_signal(self, "set-sidebar-id", view.sidebar_id)
                    emit_signal(self, "set-selection-ids", view.selection_ids)
                if visible_child is not None:
                    visible_child.stop()
                    visible_child.destroy_later()

    def remove(self, view):
        """
            Remove from stack and history
            @param view as View
        """
        if self.__history.exists(view):
            self.__history.remove(view)
        Gtk.Stack.remove(self, view)

    def save_history(self):
        """
            Save history to disk
        """
        visible_child = self.get_visible_child()
        if visible_child is not None:
            self.__history.add_view(visible_child)
        self.__history.save()

    def load_history(self):
        """
            Load history from disk
        """
        try:
            self.__history.load()
            view = self.__history.pop()
            if view is not None:
                self.add(view)
                Gtk.Stack.set_visible_child(self, view)
                if view.sidebar_id is not None:
                    emit_signal(self, "set-sidebar-id", view.sidebar_id)
                    emit_signal(self, "set-selection-ids", view.selection_ids)
        except Exception as e:
            Logger.error("AdaptiveStack::load_history(): %s", e)

    def clear(self):
        """
            Clear stack
        """
        for child in self.get_children():
            child.destroy()
        self.__history.clear()

    @property
    def history(self):
        """
            Get stack history
            @return [AdaptiveView]
        """
        return self.__history



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio, Gtk

import itertools
from time import time
from gettext import gettext as _

from lollypop.sqlcursor import SqlCursor
from lollypop.utils import translate_artist_name
from lollypop.database_history import History
from lollypop.define import App, Type, StorageType, LOLLYPOP_DATA_PATH
from lollypop.logger import Logger
from lollypop.helper_task import TaskHelper


class DatabaseUpgrade:
    """
        Manage database schema upgrades
    """

    def __init__(self):
        """
            Init object
        """
        # Here are schema upgrade, key is database version,
        # value is sql request
        self._UPGRADES = {
        }

    def upgrade(self, db):
        """
            Upgrade db
            @param db as Database
        """
        version = 0
        SqlCursor.add(db)
        with SqlCursor(db, True) as sql:
            result = sql.execute("PRAGMA user_version")
            v = result.fetchone()
            if v is not None:
                version = v[0]
            if version < self.version:
                for i in range(version + 1, self.version + 1):
                    try:
                        if isinstance(self._UPGRADES[i], str):
                            sql.execute(self._UPGRADES[i])
                            SqlCursor.commit(db)
                        else:
                            self._UPGRADES[i](db)
                            SqlCursor.commit(db)
                    except Exception as e:
                        Logger.error("DB upgrade %s failed: %s" %
                                     (i, e))
                sql.execute("PRAGMA user_version=%s" % self.version)
        SqlCursor.remove(db)

    @property
    def version(self):
        """
            Current wanted version
        """
        return len(self._UPGRADES)


class DatabasePlaylistsUpgrade(DatabaseUpgrade):
    """
        Manage database schema upgrades
    """

    def __init__(self):
        """
            Init upgrade
        """
        DatabaseUpgrade.__init__(self)
        self._UPGRADES = {
           1: "ALTER TABLE playlists ADD synced INT NOT NULL DEFAULT 0",
           2: "ALTER TABLE playlists ADD smart_enabled INT NOT NULL DEFAULT 0",
           3: "ALTER TABLE playlists ADD smart_sql TEXT",
           4: self.__upgrade_4,
           5: "ALTER TABLE playlists ADD uri TEXT"
        }

#######################
# PRIVATE             #
#######################
    def __upgrade_4(self, db):
        """
            Import tracks from loved playlist to DB
        """
        with SqlCursor(db, True) as sql1:
            result = sql1.execute("SELECT uri\
                                   FROM tracks\
                                   WHERE playlist_id=?", (Type.LOVED,))
            with SqlCursor(App().db, True) as sql2:
                for uri in list(itertools.chain(*result)):
                    sql2.execute("UPDATE tracks SET loved=1 WHERE uri=?",
                                 (uri,))


class DatabaseAlbumsUpgrade(DatabaseUpgrade):
    """
        Manage database schema upgrades
    """

    def __init__(self):
        """
            Init upgrade
        """
        DatabaseUpgrade.__init__(self)
        self._UPGRADES = {
            1: "UPDATE tracks SET duration=CAST(duration as INTEGER);",
            2: "UPDATE albums SET artist_id=-2001 where artist_id=-999;",
            3: self.__upgrade_3,
            4: self.__upgrade_4,
            5: "CREATE index idx_aa ON album_artists(album_id)",
            6: "CREATE index idx_ta ON track_artists(track_id)",
            7: "ALTER TABLE tracks ADD discname TEXT",
            8: "CREATE index idx_ag ON album_genres(album_id)",
            9: "CREATE index idx_tg ON track_genres(track_id)",
            10: "UPDATE tracks set ltime=0 where ltime is null",
            11: "ALTER TABLE albums ADD synced INT NOT NULL DEFAULT 0",
            12: "ALTER TABLE tracks ADD persistent INT NOT NULL DEFAULT 1",
            13: self.__upgrade_13,
            14: "UPDATE albums SET synced=-1 where mtime=0",
            15: self.__upgrade_15,
            16: self.__upgrade_16,
            17: "ALTER TABLE albums ADD loved INT NOT NULL DEFAULT 0",
            18: self.__upgrade_18,
            19: self.__upgrade_19,
            20: self.__upgrade_20,
            21: self.__upgrade_21,
            22: self.__upgrade_22,
            23: self.__upgrade_23,
            24: "ALTER TABLE albums ADD album_id TEXT",
            25: "ALTER TABLE tracks ADD mb_track_id TEXT",
            26: self.__upgrade_26,
            27: "UPDATE tracks SET duration=CAST(duration AS INT)",
            28: self.__upgrade_28,
            29: self.__upgrade_29,
            30: "ALTER TABLE tracks ADD loved INT NOT NULL DEFAULT 0",
            31: self.__upgrade_31,
            32: "ALTER TABLE tracks ADD bpm DOUBLE",
            33: "ALTER TABLE artists ADD mb_artist_id TEXT",
            34: self.__upgrade_31,
            35: "UPDATE albums SET synced=2 WHERE synced=1",
            36: self.__upgrade_36,
            37: self.__upgrade_37,
            38: """CREATE TABLE albums_timed_popularity (
                                                album_id INT NOT NULL,
                                                mtime INT NOT NULL,
                                                popularity INT NOT NULL)""",
            39: self.__upgrade_39,
            40: """UPDATE tracks SET duration = duration * 1000""",
            # Here we force an mb_album_id if empty, needed by artwork
            41: """UPDATE albums SET mb_album_id=rowid
                   WHERE mb_album_id is null""",
            # Fix previous update
            42: """UPDATE albums SET mb_album_id=null
                   WHERE storage_type=2 AND rowid=mb_album_id""",
            43: """CREATE TABLE featuring (artist_id INT NOT NULL,
                                           album_id INT NOT NULL)""",
            44: self.__upgrade_44,
            45: self.__upgrade_45,
            46: self.__upgrade_46,
            47: self.__upgrade_47,
            48: self.__upgrade_48,
        }

#######################
# PRIVATE             #
#######################
    def __upgrade_3(self, db):
        """
            Add a sorted field to artists
        """
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE artists ADD sortname TEXT")
            result = sql.execute("SELECT DISTINCT artists.rowid,\
                                  artists.name\
                                  FROM artists")
            for row in result:
                translated = translate_artist_name(row[1])
                sql.execute("UPDATE artists SET name=? WHERE rowid=?",
                            (translated, row[0]))
                sql.execute("UPDATE artists SET sortname=? WHERE rowid=?",
                            (row[1], row[0]))

    def __upgrade_4(self, db):
        """
            Add album artists table
        """
        with SqlCursor(db, True) as sql:
            sql.execute("CREATE TABLE album_artists (\
                                                album_id INT NOT NULL,\
                                                artist_id INT NOT NULL)")
            result = sql.execute("SELECT rowid from albums")
            for album_id in list(itertools.chain(*result)):
                result = sql.execute("SELECT artist_id\
                                     FROM albums\
                                     WHERE rowid=?",
                                     (album_id,))
                v = result.fetchone()
                if v is not None:
                    artist_id = v[0]
                    sql.execute("INSERT INTO album_artists\
                                (album_id, artist_id)\
                                VALUES(?, ?)",
                                (album_id, artist_id))
            sql.execute("CREATE TEMPORARY TABLE backup(id,\
                                                       name,\
                                                       no_album_artist,\
                                                       year,\
                                                       path,\
                                                       popularity,\
                                                       mtime)")
            sql.execute("INSERT INTO backup\
                        SELECT id,\
                               name,\
                               no_album_artist,\
                               year,\
                               path,\
                               popularity,\
                               mtime FROM albums")
            sql.execute("DROP TABLE albums")
            sql.execute("CREATE TABLE albums (id INTEGER PRIMARY KEY,\
                        name TEXT NOT NULL,\
                        no_album_artist BOOLEAN NOT NULL,\
                        year INT,\
                        path TEXT NOT NULL,\
                        popularity INT NOT NULL,\
                        mtime INT NOT NULL)")
            sql.execute("INSERT INTO albums\
                        SELECT id,\
                               name,\
                               no_album_artist,\
                               year,\
                               path,\
                               popularity,\
                               mtime FROM backup")
            sql.execute("DROP TABLE backup")

    def __upgrade_13(self, db):
        """
            Convert tracks filepath column to uri
        """
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE tracks RENAME TO tmp_tracks")
            sql.execute("""CREATE TABLE tracks (id INTEGER PRIMARY KEY,
                                              name TEXT NOT NULL,
                                              uri TEXT NOT NULL,
                                              duration INT,
                                              tracknumber INT,
                                              discnumber INT,
                                              discname TEXT,
                                              album_id INT NOT NULL,
                                              year INT,
                                              popularity INT NOT NULL,
                                              ltime INT NOT NULL,
                                              mtime INT NOT NULL,
                                              persistent INT NOT NULL
                                              DEFAULT 1)""")

            sql.execute("""INSERT INTO tracks(id, name, uri, duration,
                        tracknumber, discnumber, discname, album_id,
                        year, popularity, ltime, mtime, persistent) SELECT
                            id, name, filepath, duration,
                            tracknumber, discnumber, discname, album_id,
                            year, popularity, ltime, mtime, persistent FROM
                          tmp_tracks""")
            sql.execute("DROP TABLE tmp_tracks")
            result = sql.execute("SELECT rowid FROM tracks")
            for track_id in list(itertools.chain(*result)):
                result = sql.execute("SELECT uri FROM tracks WHERE rowid=?",
                                     (track_id,))
                v = result.fetchone()
                if v is not None:
                    uri = v[0]
                    if uri.startswith("/"):
                        uri = GLib.filename_to_uri(uri)
                        sql.execute("UPDATE tracks set uri=? WHERE rowid=?",
                                    (uri, track_id))
        with SqlCursor(App().playlists) as sql:
            sql.execute("ALTER TABLE tracks RENAME TO tmp_tracks")
            sql.execute("""CREATE TABLE tracks (playlist_id INT NOT NULL,
                                                uri TEXT NOT NULL)""")
            sql.execute("""INSERT INTO tracks(playlist_id, uri) SELECT
                            playlist_id, filepath FROM tmp_tracks""")
            sql.execute("DROP TABLE tmp_tracks")
            result = sql.execute("SELECT uri FROM tracks")
            for path in list(itertools.chain(*result)):
                if path.startswith("/"):
                    uri = GLib.filename_to_uri(path)
                    sql.execute("UPDATE tracks set uri=? WHERE uri=?",
                                (uri, path))

    def __upgrade_15(self, db):
        """
            Fix broken 0.9.208 release
        """
        pass

    def __upgrade_16(self, db):
        """
            Get rid of paths
        """
        paths = App().settings.get_value("music-path")
        uris = []
        for path in paths:
            uris.append(GLib.filename_to_uri(path))
        App().settings.set_value("music-uris", GLib.Variant("as", uris))
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE albums RENAME TO tmp_albums")
            sql.execute("""CREATE TABLE albums (
                                              id INTEGER PRIMARY KEY,
                                              name TEXT NOT NULL,
                                              no_album_artist BOOLEAN NOT NULL,
                                              year INT,
                                              uri TEXT NOT NULL,
                                              popularity INT NOT NULL,
                                              synced INT NOT NULL,
                                              mtime INT NOT NULL)""")

            sql.execute("""INSERT INTO albums(id, name, no_album_artist,
                        year, uri, popularity, synced, mtime) SELECT
                            id, name, no_album_artist,
                            year, path, popularity, synced, mtime FROM
                            tmp_albums""")
            sql.execute("DROP TABLE tmp_albums")
            result = sql.execute("SELECT rowid, uri FROM albums")
            for (rowid, uri) in result:
                if uri.startswith("/"):
                    uri = GLib.filename_to_uri(uri)
                    sql.execute("UPDATE albums set uri=? WHERE rowid=?",
                                (uri, rowid))

    def __upgrade_18(self, db):
        """
            Upgrade history
        """
        with SqlCursor(History()) as sql:
            sql.execute("ALTER TABLE history ADD loved_album\
                        INT NOT NULL DEFAULT 0")

    def __upgrade_19(self, db):
        """
            Upgrade history
        """
        with SqlCursor(History()) as sql:
            try:
                sql.execute("ALTER TABLE history ADD album_rate\
                            INT NOT NULL DEFAULT -1")
                sql.execute("ALTER TABLE history ADD rate\
                            INT NOT NULL DEFAULT -1")
            except:
                pass  # May fails if History was non existent
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE tracks ADD rate\
                        INT NOT NULL DEFAULT -1")
            sql.execute("ALTER TABLE albums ADD rate\
                        INT NOT NULL DEFAULT -1")

    def __upgrade_20(self, db):
        """
            Add mtimes tables
        """
        mtime = int(time())
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE album_genres\
                         ADD mtime INT NOT NULL DEFAULT %s" % mtime)
            sql.execute("ALTER TABLE track_genres\
                         ADD mtime INT NOT NULL DEFAULT %s" % mtime)
            # Remove mtimes from albums table
            sql.execute("CREATE TEMPORARY TABLE backup(\
                                          id INTEGER PRIMARY KEY,\
                                          name TEXT NOT NULL,\
                                          no_album_artist BOOLEAN NOT NULL,\
                                          year INT,\
                                          uri TEXT NOT NULL,\
                                          popularity INT NOT NULL,\
                                          rate INT NOT NULL,\
                                          loved INT NOT NULL,\
                                          synced INT NOT NULL)")
            sql.execute("INSERT INTO backup\
                            SELECT id,\
                                   name,\
                                   no_album_artist,\
                                   year,\
                                   uri,\
                                   popularity,\
                                   rate,\
                                   loved,\
                                   synced FROM albums")
            sql.execute("DROP TABLE albums")
            sql.execute("CREATE TABLE albums(\
                                          id INTEGER PRIMARY KEY,\
                                          name TEXT NOT NULL,\
                                          no_album_artist BOOLEAN NOT NULL,\
                                          year INT,\
                                          uri TEXT NOT NULL,\
                                          popularity INT NOT NULL,\
                                          rate INT NOT NULL,\
                                          loved INT NOT NULL,\
                                          synced INT NOT NULL)")
            sql.execute("INSERT INTO albums\
                            SELECT id,\
                                   name,\
                                   no_album_artist,\
                                   year,\
                                   uri,\
                                   popularity,\
                                   rate,\
                                   loved,\
                                   synced FROM backup")
            sql.execute("DROP TABLE backup")
            # Remove mtimes from tracks table
            sql.execute("CREATE TEMPORARY TABLE backup(\
                                          id INTEGER PRIMARY KEY,\
                                          name TEXT NOT NULL,\
                                          uri TEXT NOT NULL,\
                                          duration INT,\
                                          tracknumber INT,\
                                          discnumber INT,\
                                          discname TEXT,\
                                          album_id INT NOT NULL,\
                                          year INT,\
                                          popularity INT NOT NULL,\
                                          rate INT NOT NULL,\
                                          ltime INT NOT NULL,\
                                          persistent INT NOT NULL)")
            sql.execute("INSERT INTO backup\
                            SELECT id,\
                                   name,\
                                   uri,\
                                   duration,\
                                   tracknumber,\
                                   discnumber,\
                                   discname,\
                                   album_id,\
                                   year,\
                                   popularity,\
                                   rate,\
                                   ltime,\
                                   persistent FROM tracks")
            sql.execute("DROP TABLE tracks")
            sql.execute("CREATE TABLE tracks(\
                                          id INTEGER PRIMARY KEY,\
                                          name TEXT NOT NULL,\
                                          uri TEXT NOT NULL,\
                                          duration INT,\
                                          tracknumber INT,\
                                          discnumber INT,\
                                          discname TEXT,\
                                          album_id INT NOT NULL,\
                                          year INT,\
                                          popularity INT NOT NULL,\
                                          rate INT NOT NULL,\
                                          ltime INT NOT NULL,\
                                          persistent INT NOT NULL)")
            sql.execute("INSERT INTO tracks\
                            SELECT id,\
                                   name,\
                                   uri,\
                                   duration,\
                                   tracknumber,\
                                   discnumber,\
                                   discname,\
                                   album_id,\
                                   year,\
                                   popularity,\
                                   rate,\
                                   ltime,\
                                   persistent FROM backup")
            sql.execute("DROP TABLE backup")

    def __upgrade_21(self, db):
        pass

    def __upgrade_22(self, db):
        """
            Remove Charts/Web entries
        """
        with SqlCursor(db, True) as sql:
            # Remove persistent from tracks table
            sql.execute("CREATE TEMPORARY TABLE backup(\
                                          id INTEGER PRIMARY KEY,\
                                          name TEXT NOT NULL,\
                                          uri TEXT NOT NULL,\
                                          duration INT,\
                                          tracknumber INT,\
                                          discnumber INT,\
                                          discname TEXT,\
                                          album_id INT NOT NULL,\
                                          year INT,\
                                          popularity INT NOT NULL,\
                                          rate INT NOT NULL,\
                                          ltime INT NOT NULL)")
            sql.execute("INSERT INTO backup\
                            SELECT id,\
                                   name,\
                                   uri,\
                                   duration,\
                                   tracknumber,\
                                   discnumber,\
                                   discname,\
                                   album_id,\
                                   year,\
                                   popularity,\
                                   rate,\
                                   ltime FROM tracks")
            sql.execute("DROP TABLE tracks")
            sql.execute("CREATE TABLE tracks(\
                                          id INTEGER PRIMARY KEY,\
                                          name TEXT NOT NULL,\
                                          uri TEXT NOT NULL,\
                                          duration INT,\
                                          tracknumber INT,\
                                          discnumber INT,\
                                          discname TEXT,\
                                          album_id INT NOT NULL,\
                                          year INT,\
                                          popularity INT NOT NULL,\
                                          rate INT NOT NULL,\
                                          ltime INT NOT NULL)")
            sql.execute("INSERT INTO tracks\
                            SELECT id,\
                                   name,\
                                   uri,\
                                   duration,\
                                   tracknumber,\
                                   discnumber,\
                                   discname,\
                                   album_id,\
                                   year,\
                                   popularity,\
                                   rate,\
                                   ltime FROM backup")
            sql.execute("DROP TABLE backup")

    def __upgrade_23(self, db):
        """
            Restore back mtime in tracks
        """
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE tracks ADD mtime INT")
            sql.execute("ALTER TABLE albums ADD mtime INT")

            sql.execute("UPDATE tracks SET mtime = (\
                            SELECT mtime FROM track_genres\
                            WHERE track_genres.track_id=tracks.rowid)")

            sql.execute("UPDATE albums SET mtime = (\
                            SELECT mtime FROM album_genres\
                            WHERE album_genres.album_id=albums.rowid)")
            # Remove mtime from album_genres table
            sql.execute("CREATE TABLE album_genres2 (\
                                                album_id INT NOT NULL,\
                                                genre_id INT NOT NULL)")
            sql.execute("INSERT INTO album_genres2\
                            SELECT album_id,\
                                   genre_id FROM album_genres")
            sql.execute("DROP TABLE album_genres")
            sql.execute("ALTER TABLE album_genres2 RENAME TO album_genres")

            # Remove mtime from track_genres table
            sql.execute("CREATE TABLE track_genres2 (\
                                                track_id INT NOT NULL,\
                                                genre_id INT NOT NULL)")
            sql.execute("INSERT INTO track_genres2\
                            SELECT track_id,\
                                   genre_id FROM track_genres")
            sql.execute("DROP TABLE track_genres")
            sql.execute("ALTER TABLE track_genres2 RENAME TO track_genres")

    def __upgrade_26(self, db):
        """
            Rename album_id to mb_album_id in albums
        """
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE albums RENAME TO tmp_albums")
            sql.execute("""CREATE TABLE albums (
                               id INTEGER PRIMARY KEY,
                               name TEXT NOT NULL,
                               mb_album_id TEXT,
                               no_album_artist BOOLEAN NOT NULL,
                               year INT,
                               uri TEXT NOT NULL,
                               popularity INT NOT NULL,
                               rate INT NOT NULL,
                               loved INT NOT NULL,
                               mtime INT NOT NULL,
                               synced INT NOT NULL)""")

            sql.execute("""INSERT INTO albums (id, name, mb_album_id,
                            no_album_artist, year, uri, popularity, rate,
                            loved, mtime, synced) SELECT id, name, album_id,
                            no_album_artist, year, uri, popularity, rate,
                            loved, mtime, synced FROM tmp_albums""")
            sql.execute("DROP TABLE tmp_albums")

    def __upgrade_28(self, db):
        """
            Upgrade setting based on db
            https://gitlab.gnome.org/gnumdk/lollypop/issues/1368
        """
        with SqlCursor(db, True) as sql:
            result = sql.execute("SELECT albums.rowid\
                                  FROM albums, album_artists\
                                  WHERE album_artists.artist_id=?\
                                  AND album_artists.album_id=albums.rowid\
                                  LIMIT 1",
                                 (Type.COMPILATIONS,))
            if list(itertools.chain(*result)):
                App().settings.set_value("show-compilations",
                                         GLib.Variant("b", True))

    def __upgrade_29(self, db):
        """
            Upgrade year to year
        """
        from time import strptime, mktime
        from datetime import datetime
        for item in ["albums", "tracks"]:
            with SqlCursor(db, True) as sql:
                sql.execute("ALTER TABLE %s ADD timestamp INT" % item)
                result = sql.execute("SELECT rowid, year FROM %s" % item)
                for (rowid, year) in result:
                    if year is None:
                        continue
                    elif len(str(year)) == 2:
                        struct = strptime(str(year), "%y")
                    elif len(str(year)) == 4:
                        struct = strptime(str(year), "%Y")
                    else:
                        continue
                    dt = datetime.fromtimestamp(mktime(struct))
                    timestamp = dt.timestamp()
                    sql.execute(
                        "UPDATE %s set timestamp=? WHERE rowid=?" % item,
                        (timestamp, rowid))

    def __upgrade_31(self, db):
        """
            Delete history database related to upgrade 30
        """
        try:
            LOCAL_PATH = GLib.get_user_data_dir() + "/lollypop"
            DB_PATH = "%s/history.db" % LOCAL_PATH
            f = Gio.File.new_for_path(DB_PATH)
            f.delete(None)
        except Exception as e:
            Logger.error("DatabaseAlbumsUpgrade::__upgrade_31(): %s", e)

    def __upgrade_36(self, db):
        """
            Restore back mtime in tracks
        """
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE tracks ADD storage_type INT")
            sql.execute("ALTER TABLE albums ADD storage_type INT")
            sql.execute("UPDATE tracks SET storage_type=?\
                         WHERE mtime > 0", (StorageType.COLLECTION,))
            sql.execute("UPDATE albums SET storage_type=?\
                         WHERE mtime > 0", (StorageType.COLLECTION,))
            sql.execute("UPDATE tracks SET storage_type=?\
                         WHERE mtime = -1", (StorageType.SAVED,))
            sql.execute("UPDATE albums SET storage_type=?\
                         WHERE mtime = -1", (StorageType.SAVED,))

    def __upgrade_37(self, db):
        """
            Update Type.WEB and Type.COMPILATIONS
        """
        App().settings.reset("shown-album-lists")
        with SqlCursor(db, True) as sql:
            sql.execute("UPDATE track_genres SET genre_id=-9\
                         WHERE genre_id=-22")
            sql.execute("UPDATE album_genres SET genre_id=-9\
                         WHERE genre_id=-22")
            sql.execute("UPDATE album_artists SET artist_id=-10\
                         WHERE artist_id=-2001")

    def __upgrade_39(self, db):
        """
            Reset Spotify tracks: we are now using Spotify id as MusicBrainz id
        """
        with SqlCursor(db, True) as sql:
            for storage_type in [StorageType.SPOTIFY_NEW_RELEASES,
                                 StorageType.SPOTIFY_SIMILARS]:
                sql.execute("DELETE FROM tracks WHERE\
                             storage_type=? AND mb_track_id=''",
                            (storage_type,))
            sql.execute("DELETE FROM track_artists\
                         WHERE track_artists.track_id NOT IN (\
                            SELECT tracks.rowid FROM tracks)")
            sql.execute("DELETE FROM track_genres\
                         WHERE track_genres.track_id NOT IN (\
                            SELECT tracks.rowid FROM tracks)")
            sql.execute("DELETE FROM albums WHERE albums.rowid NOT IN (\
                            SELECT tracks.album_id FROM tracks)")
            sql.execute("DELETE FROM album_genres\
                         WHERE album_genres.album_id NOT IN (\
                            SELECT albums.rowid FROM albums)")
            sql.execute("DELETE FROM album_artists\
                         WHERE album_artists.album_id NOT IN (\
                            SELECT albums.rowid FROM albums)")
            sql.execute("DELETE FROM albums_timed_popularity\
                         WHERE albums_timed_popularity.album_id NOT IN (\
                            SELECT albums.rowid FROM albums)")
            sql.execute("DELETE FROM artists WHERE artists.rowid NOT IN (\
                            SELECT album_artists.artist_id\
                            FROM album_artists) AND artists.rowid NOT IN (\
                                SELECT track_artists.artist_id\
                                FROM track_artists)")

    def __upgrade_44(self, db):
        """
            Delete spotify albums as spotify id is not stored in URI
        """
        from lollypop.database_albums import AlbumsDatabase
        from lollypop.database_artists import ArtistsDatabase
        from lollypop.database_tracks import TracksDatabase
        albums = AlbumsDatabase(db)
        artists = ArtistsDatabase(db)
        tracks = TracksDatabase(db)
        for storage_type in [StorageType.SPOTIFY_NEW_RELEASES,
                             StorageType.SPOTIFY_SIMILARS,
                             StorageType.DEEZER_CHARTS]:
            album_ids = albums.get_for_storage_type(storage_type)
            for album_id in album_ids:
                # EPHEMERAL with not tracks will be cleaned below
                albums.set_storage_type(album_id, StorageType.EPHEMERAL)
                tracks.remove_album(album_id)
        tracks.clean()
        albums.clean()
        artists.clean()

    def __upgrade_45(self, db):
        """
            Add lp_album_id/lp_track_id
        """
        with SqlCursor(db, True) as sql:
            sql.execute("ALTER TABLE tracks ADD lp_track_id TEXT")
            sql.execute("ALTER TABLE albums ADD lp_album_id TEXT")

    def __upgrade_46(self, db):
        """
            Populate lp_album_id/lp_track_id
        """
        queue = LOLLYPOP_DATA_PATH + "/queue.bin"
        try:
            f = Gio.File.new_for_path(queue)
            f.delete(None)
        except:
            pass
        from lollypop.database_albums import AlbumsDatabase
        from lollypop.database_tracks import TracksDatabase
        from lollypop.utils import get_lollypop_album_id, get_lollypop_track_id
        albums = AlbumsDatabase(db)
        tracks = TracksDatabase(db)

        def do_migration(dialog, label, progress):
            GLib.idle_add(
                label.set_text,
                _("Please wait while Lollypop is updating albums"))
            album_ids = albums.get_ids([], [], StorageType.ALL, True)
            album_ids += albums.get_compilation_ids([], StorageType.ALL, True)
            count = len(album_ids)
            i = 0
            for album_id in album_ids:
                if i % 10 == 0:
                    GLib.idle_add(progress.set_fraction, i / count)
                name = albums.get_name(album_id)
                artists = ";".join(albums.get_artists(album_id))
                lp_album_id = get_lollypop_album_id(name, artists, None)
                albums.set_lp_album_id(album_id, lp_album_id)
                i += 1

            track_ids = tracks.get_ids(StorageType.ALL, True)
            count = len(track_ids)
            i = 0
            GLib.idle_add(
                label.set_text,
                _("Please wait while Lollypop is updating tracks"))
            for track_id in track_ids:
                if i % 10 == 0:
                    GLib.idle_add(progress.set_fraction, i / count)
                name = tracks.get_name(track_id)
                artists = ";".join(tracks.get_artists(track_id))
                album_name = tracks.get_album_name(track_id)
                lp_track_id = get_lollypop_track_id(name, artists, album_name)
                tracks.set_lp_track_id(track_id, lp_track_id)
                i += 1
            GLib.idle_add(dialog.destroy)

        dialog = Gtk.MessageDialog(buttons=Gtk.ButtonsType.NONE)
        progress = Gtk.ProgressBar.new()
        progress.show()
        label = Gtk.Label.new()
        label.show()
        grid = Gtk.Grid.new()
        grid.set_orientation(Gtk.Orientation.VERTICAL)
        grid.set_row_spacing(10)
        grid.show()
        grid.add(label)
        grid.add(progress)
        dialog.set_image(grid)
        helper = TaskHelper()
        helper.run(do_migration, dialog, label, progress)
        dialog.run()

    def __upgrade_47(self, db):
        """
        Update the covers for playlists.
        """
        from lollypop.art import clean_all_cache
        clean_all_cache()

    def __upgrade_48(self, db):
        """
            Convert loved to new flags
        """
        with SqlCursor(db, True) as sql:
            sql.execute("UPDATE tracks set loved=2 where loved=1")
            sql.execute("UPDATE tracks set loved=1 where loved=0")
            sql.execute("UPDATE tracks set loved=4 where loved=-1")
            sql.execute("UPDATE albums set loved=2 where loved=1")
            sql.execute("UPDATE albums set loved=1 where loved=0")
            sql.execute("UPDATE albums set loved=4 where loved=-1")



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from lollypop.define import App, ArtBehaviour
from lollypop.objects_album import Album
from lollypop.helper_signals import SignalsHelper, signals


class ArtworkPlayerWidget(Gtk.Image, SignalsHelper):
    """
        Gtk.Image auto updated with current player state
    """

    @signals
    def __init__(self, behaviour=ArtBehaviour.CACHE):
        """
            Init artwork
            @param behaviour as ArtBehaviour
        """
        Gtk.Image.__init__(self)
        self.__behaviour = behaviour
        self.__width = self.__height = 1
        self.__previous_artwork_id = None
        self.set_property("valign", Gtk.Align.CENTER)
        self.set_property("halign", Gtk.Align.CENTER)
        self.__per_track_cover = App().settings.get_value(
            "allow-per-track-cover")
        return [
            (App().player, "current-changed", "_on_current_changed"),
            (App().album_art, "album-artwork-changed",
             "_on_album_artwork_changed"),
        ]

    def set_art_size(self, width, height):
        """
            Set artwork width/height
            @param width as int
            @param heigth as int
        """
        self.__width = width
        self.__height = height

    def update(self, force=False):
        """
            Update artwork
            @param force as bool
        """
        same_artwork = not force and self.__previous_artwork_id ==\
            App().player.current_track.album.id and not self.__per_track_cover
        if same_artwork:
            return
        self.__previous_artwork_id = App().player.current_track.album.id
        self.set_artwork(self.__width, self.__height,
                         self.__on_artwork, self.__behaviour)

    def set_behaviour(self, behaviour):
        """
            Set artwork behaviour
            @param behaviour as ArtBehaviour
        """
        self.__behaviour = behaviour
        self.__previous_artwork_id = None
        context_style = self.get_style_context()
        if self.__behaviour & ArtBehaviour.ROUNDED:
            context_style.add_class("rounded")
        else:
            context_style.remove_class("rounded")

    def set_artwork(self, width, height, callback, behaviour):
        """
            Set artwork
            @param width as int
            @param height as int
            @param callback as function
        """
        scale_factor = self.get_scale_factor()
        if App().player.current_track.id is not None:
            if self.__per_track_cover:
                behaviour |= ArtBehaviour.NO_CACHE
                album = Album(App().player.current_track.album.id)
                App().album_art.uncache(album,
                                        width * scale_factor,
                                        height * scale_factor)
                album.set_tracks([App().player.current_track])
            else:
                album = App().player.current_track.album
            App().art_helper.set_album_artwork(
                album,
                width,
                height,
                scale_factor,
                behaviour,
                callback)
        else:
            self.set_from_surface(None)

#######################
# PROTECTED           #
#######################
    def _on_current_changed(self, player):
        """
            Update artwork and labels
            @param player as Player
        """
        self.update()

    def _on_album_artwork_changed(self, art, album_id):
        """
            Update cover for album_id
            @param art as Art
            @param album_id as int
        """
        if App().player.current_track.album.id == album_id:
            self.set_artwork(self.__width, self.__height,
                             self.__on_artwork, self.__behaviour)

#######################
# PRIVATE             #
#######################
    def __on_artwork(self, surface):
        """
            Set artwork
            @param surface as str
        """
        self.set_size_request(self.__width, self.__height)
        context_style = self.get_style_context()
        if surface is None:
            context_style.add_class("cover-background")
            if self.__behaviour & ArtBehaviour.ROUNDED:
                context_style.add_class("rounded")
            icon_name = "folder-music-symbolic"
            self.set_from_icon_name(icon_name, Gtk.IconSize.INVALID)
            self.set_pixel_size(self.__width / 3)
        else:
            context_style.remove_class("cover-background")
            if self.__behaviour & ArtBehaviour.ROUNDED:
                context_style.remove_class("rounded")
            self.set_from_surface(surface)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GdkPixbuf, GLib

from hashlib import md5

from lollypop.artwork_manager import ArtworkManager
from lollypop.artwork_downloader_artist import ArtistArtworkDownloader
from lollypop.logger import Logger
from lollypop.define import CACHE_PATH
from lollypop.define import ARTISTS_PATH, ArtBehaviour, ArtSize
from lollypop.define import StoreExtention
from lollypop.utils import emit_signal
from lollypop.utils_file import create_dir


class ArtistArtwork(ArtworkManager, ArtistArtworkDownloader):
    """
        Artist artwork manager
    """

    def __init__(self):
        """
            Init artist artwork manager
        """
        ArtworkManager.__init__(self)
        ArtistArtworkDownloader.__init__(self)
        create_dir(ARTISTS_PATH)

    def get_path(self, artist):
        """
            Get artwork path for artist
            @param artist as str
            @return str/None
        """
        encoded = self.__encode(artist)
        if self.extension == StoreExtention.PNG:
            extensions = ["png", "jpg"]
        else:
            extensions = ["jpg", "png"]
        for extension in extensions:
            cache_path = "%s/%s.%s" % (ARTISTS_PATH, encoded, extension)
            if GLib.file_test(cache_path, GLib.FileTest.EXISTS):
                return cache_path
        return None

    def add(self, artist, data, storage_type):
        """
            Add artist artwork to store
            @param artist as str
            @param data as bytes
            @param storage_type as StorageType
            @thread safe
        """
        self.__uncache(artist)
        encoded = self.__encode(artist)
        cache_path = "%s/%s" % (ARTISTS_PATH, encoded)
        cache_path = self.add_extension(cache_path)
        self.save_pixbuf_from_data(cache_path, data)
        emit_signal(self, "artist-artwork-changed", artist)

    def get(self, artist, width, height, scale_factor,
            behaviour=ArtBehaviour.CACHE):
        """
            Return a cairo surface for album_id
            @param artist as str
            @param width as int
            @param height as int
            @param scale_factor as int
            @param behaviour as ArtBehaviour
            @return cairo surface
            @thread safe
        """
        width *= scale_factor
        height *= scale_factor
        # Blur when reading from tags can be slow, so prefer cached version
        # Blur allows us to ignore width/height until we want CROP/CACHE
        optimized_blur = behaviour & (ArtBehaviour.BLUR |
                                      ArtBehaviour.BLUR_HARD) and\
            not behaviour & (ArtBehaviour.CACHE |
                             ArtBehaviour.CROP |
                             ArtBehaviour.CROP_SQUARE)
        if optimized_blur:
            w = ArtSize.BANNER * scale_factor
            h = ArtSize.BANNER * scale_factor
        else:
            w = width
            h = height
        filename = self.__encode(artist)
        cache_path = "%s/%s_%s_%s" % (CACHE_PATH, filename, w, h)
        cache_path = self.add_extension(cache_path)
        pixbuf = None
        try:
            # Look in cache
            f = Gio.File.new_for_path(cache_path)
            if not behaviour & ArtBehaviour.NO_CACHE and f.query_exists():
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(cache_path)
                if optimized_blur:
                    pixbuf = self.load_behaviour(pixbuf,
                                                 width, height, behaviour)
                return pixbuf
            else:
                artwork_path = self.get_path(artist)
                if artwork_path is not None:
                    pixbuf = GdkPixbuf.Pixbuf.new_from_file(artwork_path)
                else:
                    self.download(artist)
                    return None
                pixbuf = self.load_behaviour(pixbuf,
                                             width, height, behaviour)
                if behaviour & ArtBehaviour.CACHE:
                    self.save_pixbuf(pixbuf, cache_path)
            return pixbuf
        except Exception as e:
            Logger.warning("ArtistArtwork::get(): %s" % e)
            return None

#######################
# PRIVATE             #
#######################
    def __encode(self, artist):
        """
            Get a uniq string for artist
            @param artist as str
        """
        return md5(artist.encode("utf-8")).hexdigest()

    def __uncache(self, artist):
        """
            Remove artwork from cache
            @param artist as str
        """
        try:
            from pathlib import Path
            if self.extension == StoreExtention.PNG:
                extension = "png"
            else:
                extension = "jpg"
            search = "%s*.%s" % (self.__encode(artist), extension)
            for p in Path(CACHE_PATH).glob(search):
                p.unlink()
        except Exception as e:
            Logger.error("ArtistArtwork::__uncache(): %s" % e)



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


class JamendoWebHelper:
    """
        Web helper for Jamendo
    """

    def __init__(self):
        """
            Init helper
        """
        pass

    def lollypop_album_payload(self, payload):
        """
            Convert payload to Lollypop one
            @param payload as {}
            return {}
        """
        lollypop_payload = {}
        lollypop_payload["mbid"] = None
        lollypop_payload["name"] = payload["name"]
        lollypop_payload["uri"] = "jam:%s" % payload["id"]
        lollypop_payload["artists"] = payload["artist_name"]
        lollypop_payload["track-count"] = -1
        lollypop_payload["artwork-uri"] = payload["image"]
        try:
            lollypop_payload["date"] = "%sT00:00:00" %\
                payload["releasedate"]
        except:
            lollypop_payload["date"] = None
        return lollypop_payload

    def lollypop_track_payload(self, payload):
        """
            Convert payload to Lollypop one
            @param payload as {}
            @return {}
        """
        lollypop_payload = {}
        lollypop_payload["mbid"] = None
        lollypop_payload["name"] = payload["name"]
        lollypop_payload["uri"] = payload["audio"]
        lollypop_payload["artists"] = payload["artist_name"]
        lollypop_payload["discnumber"] = 1
        lollypop_payload["tracknumber"] = payload["position"]
        lollypop_payload["duration"] = int(payload["duration"]) * 1000
        return lollypop_payload

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.define import App, SelectionListMask, Type, ScanUpdate
from lollypop.utils import get_default_storage_type


class ScannerContainer:
    """
        Scanner management for main view
    """

    def __init__(self):
        """
            Init container
        """
        App().scanner.connect("updated", self.__on_collection_updated)

############
# PRIVATE  #
############
    def __handle_genre_update(self, genre_id, scan_update):
        """
            Add genre to genre list
            @param genre_id as int
            @param scan_update as ScanUpdate
        """
        if Type.GENRES_LIST in self.sidebar.selected_ids:
            storage_type = get_default_storage_type()
            if scan_update == ScanUpdate.ADDED:
                genre_name = App().genres.get_name(genre_id)
                self.left_list.add_value((genre_id, genre_name, genre_name))
            elif not App().artists.get_ids([genre_id], storage_type):
                self.left_list.remove_value(genre_id)

    def __handle_artist_update(self, artist_id, scan_update):
        """
            Add/remove artist to/from list
            @param artist_id as int
            @param scan_update as ScanUpdate
        """
        if Type.GENRES_LIST in self.sidebar.selected_ids:
            selection_list = self.right_list
            genre_ids = self.left_list.selected_ids
        elif Type.ARTISTS_LIST in self.sidebar.selected_ids and\
                self.left_list.mask & SelectionListMask.ARTISTS:
            selection_list = self.left_list
            genre_ids = []
        else:
            return
        storage_type = get_default_storage_type()
        artist_ids = App().artists.get_ids(genre_ids, storage_type)
        # We only test add, remove and absent is safe
        if artist_id not in artist_ids and scan_update == ScanUpdate.ADDED:
            return
        artist_name = App().artists.get_name(artist_id)
        sortname = App().artists.get_sortname(artist_id)
        genre_ids = []
        if scan_update == ScanUpdate.ADDED:
            selection_list.add_value((artist_id, artist_name, sortname))
        elif not App().albums.get_ids(
                genre_ids, [artist_id], storage_type, True):
            selection_list.remove_value(artist_id)

    def __on_collection_updated(self, scanner, item, scan_update):
        """
            Update lists based on collection changes
            @param scanner as CollectionScanner
            @param item as CollectionItem
            @param scan_update as ScanUpdate
        """
        for genre_id in item.genre_ids:
            self.__handle_genre_update(genre_id, scan_update)
        for artist_id in item.artist_ids:
            self.__handle_artist_update(artist_id, scan_update)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

import sqlite3
from threading import Lock

from lollypop.define import CACHE_PATH
from lollypop.sqlcursor import SqlCursor
from lollypop.database import Database
from lollypop.logger import Logger


class CacheDatabase:
    """
        Cache calculation into database
    """
    DB_PATH = "%s/cache_v1.db" % CACHE_PATH

    # SQLite documentation:
    # In SQLite, a column with type INTEGER PRIMARY KEY
    # is an alias for the ROWID.
    # Here, we define an id INT PRIMARY KEY but never feed it,
    # this make VACUUM not destroy rowids...
    __create_duration = """CREATE TABLE duration (
                            id TEXT PRIMARY KEY,
                            album_id INT NOT NULL,
                            duration INT NOT NULL DEFAULT 0)"""

    def __init__(self):
        """
            Create database tables
        """
        self.thread_lock = Lock()
        f = Gio.File.new_for_path(self.DB_PATH)
        if not f.query_exists():
            try:
                d = Gio.File.new_for_path(CACHE_PATH)
                if not d.query_exists():
                    d.make_directory_with_parents()
                # Create db schema
                with SqlCursor(self, True) as sql:
                    sql.execute(self.__create_duration)
            except Exception as e:
                Logger.error("DatabaseCache::__init__(): %s" % e)

    def set_duration(self, album_id, album_hash, duration):
        """
            Set duration in cache for album
            @param album_id as int
            @param album_hash as str
            @param duration as int
        """
        try:
            with SqlCursor(self, True) as sql:
                sql.execute("INSERT INTO duration (id, album_id, duration)\
                             VALUES (?, ?, ?)",
                            (album_hash, album_id, duration))
        except Exception as e:
            Logger.error("DatabaseCache::set_duration(): %s", e)

    def get_duration(self, album_hash):
        """
            Get duration in cache for album
            @param album_hash as str
            @return int/None
        """
        try:
            with SqlCursor(self) as sql:
                result = sql.execute("SELECT duration\
                                      FROM duration\
                                      WHERE id=?",
                                     (album_hash,))
                v = result.fetchone()
                if v is not None:
                    return v[0]
        except Exception as e:
            Logger.error("DatabaseCache::get_duration(): %s", e)
        return None

    def clear_durations(self, album_id):
        """
            Clear durations for album_id
            @param album_id as int
        """
        with SqlCursor(self, True) as sql:
            sql.execute("DELETE FROM duration WHERE album_id=?",
                        (album_id,))

    def clear_table(self, table):
        """
            Clear table
            @param table as str
        """
        with SqlCursor(self, True) as sql:
            sql.execute('DELETE FROM "%s"' % table)

    def clean(self, commit=True):
        """
            Clean cache
            @param commit as bool
        """
        with SqlCursor(self, commit) as sql:
            sql.execute('ATTACH DATABASE "%s" AS music' % Database.DB_PATH)
            sql.execute("DELETE FROM duration WHERE duration.album_id NOT IN (\
                            SELECT albums.rowid FROM music.albums)")

    def get_cursor(self):
        """
            Return a new sqlite cursor
        """
        try:
            c = sqlite3.connect(self.DB_PATH, 600.0)
            return c
        except:
            exit(-1)

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, GObject, GLib

from lollypop.define import App
from lollypop.utils import is_device, emit_signal
from lollypop.widgets_popover import Popover


class DevicesPopover(Popover):
    """
        Popover with connected devices
    """

    __gsignals__ = {
        "content-changed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
    }

    def __init__(self, progressbar):
        """
            Init popover
            @param progressbar Gtk.ProgressBar
        """
        Popover.__init__(self, False)
        self.__syncing = 0
        self.__timeout_id = None
        self.__progressbar = progressbar
        self.__scrolled = Gtk.ScrolledWindow()
        self.__scrolled.set_policy(Gtk.PolicyType.NEVER,
                                   Gtk.PolicyType.AUTOMATIC)
        self.__scrolled.show()
        self.__viewport = Gtk.Viewport()
        self.__viewport.set_margin_start(5)
        self.__viewport.set_margin_end(5)
        self.__viewport.set_margin_top(5)
        self.__viewport.set_margin_bottom(5)
        self.__scrolled.add(self.__viewport)
        self.__viewport.show()
        self.__listbox = Gtk.ListBox()
        self.__listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        # Volume manager
        self.__vm = Gio.VolumeMonitor.get()
        self.__vm.connect("mount-added", self.__on_mount_added)
        self.__vm.connect("mount-removed", self.__on_mount_removed)
        self.__viewport.add(self.__listbox)
        self.add(self.__scrolled)

    def populate(self):
        """
            Populate widget with current available devices
        """
        for mount in self.__vm.get_mounts():
            self.__add_device(mount)

    def add_fake_phone(self):
        """
            Add a fake phone device
        """
        from lollypop.widgets_device import DeviceWidget
        name = "Librem phone"
        uri = "file:///tmp/librem/"
        d = Gio.File.new_for_uri(uri + "Internal Memory")
        if not d.query_exists():
            d.make_directory_with_parents()
        d = Gio.File.new_for_uri(uri + "SD Card")
        if not d.query_exists():
            d.make_directory_with_parents()
        widget = DeviceWidget(name, uri)
        widget.connect("syncing", self.__on_syncing)
        widget.connect("size-allocate", self.__on_size_allocate)
        widget.show()
        self.__listbox.add(widget)
        self.__listbox.show()
        emit_signal(self, "content-changed",
                    len(self.__listbox.get_children()))

    @property
    def devices(self):
        """
            Get current devices
            @return [str]
        """
        return [child.name for child in self.__listbox.get_children()]

#######################
# PROTECTED           #
#######################
    def _on_button_clicked(self, button):
        """
            Hide popover
            @param button as Gtk.Button
        """
        self.hide()

    def _on_volume_value_changed(self, scale):
        """
            Set volume
            @param scale as Gtk.Scale
        """
        new_volume = scale.get_value()
        if new_volume != App().player.volume:
            App().player.set_volume(scale.get_value())

#######################
# PRIVATE             #
#######################
    def __add_device(self, mount):
        """
            Add a device
            @param mount as Gio.Mount
        """
        if is_device(mount):
            from lollypop.widgets_device import DeviceWidget
            name = mount.get_name()
            uri = mount.get_default_location().get_uri()
            if mount.get_volume() is not None:
                icon = mount.get_volume().get_symbolic_icon()
            else:
                icon = None
            widget = DeviceWidget(name, uri, icon)
            widget.connect("syncing", self.__on_syncing)
            widget.connect("size-allocate", self.__on_size_allocate)
            widget.show()
            self.__listbox.add(widget)
            self.__listbox.show()
        emit_signal(self, "content-changed",
                    len(self.__listbox.get_children()))

    def __remove_device(self, mount):
        """
            Remove volume from device list
            @param mount as Gio.Mount
        """
        uri = mount.get_default_location().get_uri()
        for widget in self.__listbox.get_children():
            if widget.uri == uri:
                widget.disconnect_by_func(self.__on_syncing)
                widget.destroy()
        emit_signal(self, "content-changed",
                    len(self.__listbox.get_children()))

    def __update_progress(self):
        """
            Update progressbar
        """
        progress = 0.0
        nb_syncs = 0
        for row in self.__listbox.get_children():
            nb_syncs += 1
            progress += row.progress
        if nb_syncs:
            value = progress / nb_syncs
            self.__progressbar.set_fraction(value)
        return True

    def __on_mount_added(self, vm, mount):
        """
            On volume mounter
            @param vm as Gio.VolumeMonitor
            @param mount as Gio.Mount
        """
        self.__add_device(mount)

    def __on_mount_removed(self, vm, mount):
        """
            On volume removed, clean selection list
            @param vm as Gio.VolumeMonitor
            @param mount as Gio.Mount
        """
        self.__remove_device(mount)

    def __on_size_allocate(self, widget, allocation):
        """
            Show scrollbar if needed
            @param widget as Gtk.Widget
            @param allocation as Gtk.Allocation
        """
        height = 10
        for child in self.__listbox.get_children():
            if child.is_visible():
                height += child.get_allocated_height()
        self.__scrolled.set_size_request(300, min(400, height))

    def __on_syncing(self, widget, status):
        """
            Start/stop progress status
            @param widget as Gtk.Widget
            @param status as bool
        """
        def hide_progress():
            if self.__timeout_id is None:
                self.__progressbar.hide()

        if status:
            self.__syncing += 1
        else:
            self.__syncing -= 1
        if self.__syncing > 0 and self.__timeout_id is None:
            self.__progressbar.set_fraction(0)
            self.__progressbar.show()
            self.__timeout_id = GLib.timeout_add(1000,
                                                 self.__update_progress)
        elif self.__syncing == 0 and self.__timeout_id is not None:
            GLib.timeout_add(1000, hide_progress)
            GLib.source_remove(self.__timeout_id)
            self.__timeout_id = None
            self.__progressbar.set_fraction(1)
